import java.io.DataOutputStream;
import java.io.IOException;
import java.net.DatagramSocket;
import java.net.DatagramPacket;
import java.net.InetAddress;
import java.net.Socket;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.Random;
import java.util.Stack;
import java.util.logging.Logger;
import java.util.logging.Level;

public class ReflectorClient {
	private static Logger logger
		= Logger.getLogger(ReflectorClient.class.getName());
	private int channel;
	private int id;
	private long lastAccessTime;

	private DatagramSocket dSocket;
	private InetAddress udpAddress;
	private int udpPort;
	private Socket socket;
	private DataOutputStream dout;

	public ReflectorClient(int channel, int id, DatagramSocket dsock,
			InetAddress address, int port) {
		this.channel = channel;
		this.id = id;
		this.lastAccessTime = System.currentTimeMillis();
		this.dSocket = dsock;
		this.udpAddress = address;
		this.udpPort = port;
		this.socket = null;
	}

	public ReflectorClient(int channel, int id, DatagramSocket dsock,
			Socket socket) {
		this.channel = channel;
		this.id = id;
		this.lastAccessTime = System.currentTimeMillis();
		this.dSocket = dsock;
		this.udpAddress = null;
		this.udpPort = -1;
		this.socket = socket;
	}

	public int getChannel() {
		return channel;
	}

	public int getId() {
		return id;
	}

	public long getLastAccessTime() {
		return lastAccessTime;
	}

	public DatagramSocket getDatagramSocket() {
		return dSocket;
	}

	public InetAddress getUDPAddress() {
		return udpAddress;
	}

	public int getUDPPort() {
		return udpPort;
	}

	public Socket getSocket() {
		return socket;
	}

	public synchronized boolean isOwnPacket(ReflectorPacket packet) {
		if (channel != packet.getChannel() || id != packet.getId())
			return false;
		if (packet.getSocket() != null) {
			// TCP packet
			if (socket == null) {
				if (udpAddress != null && !udpAddress.equals(
						packet.getSocket().getInetAddress()))
					return false;
				socket = packet.getSocket();
				lastAccessTime = System.currentTimeMillis();
				return true;
			} else {
				if (socket == packet.getSocket()) {
					lastAccessTime = System.currentTimeMillis();
					return true;
				} else {
					return false;
				}
			}
		} else {
			// UDP packet
			if (udpAddress == null) {
				if (socket != null && !socket.getInetAddress().equals(
						packet.getUDPAddress()))
					return false;
				udpAddress = packet.getUDPAddress();
				udpPort = packet.getUDPPort();
				lastAccessTime = System.currentTimeMillis();
				return true;
			} else {
				if ((udpAddress.equals(packet.getUDPAddress())
						&& udpPort == packet.getUDPPort())) {
					lastAccessTime = System.currentTimeMillis();
					return true;
				} else {
					return false;
				}
			}
		}
	}

	private synchronized void sendTCP(ReflectorPacket packet)
			throws IOException {
		byte[] buf = packet.getBuffer();
		try {
			if (dout == null)
				dout = new DataOutputStream(socket.getOutputStream());
			dout.writeShort(buf.length);
			dout.write(buf, 0, buf.length);
			dout.flush();
		} catch (IOException ex) {
			try {
				dout.close();
			} catch (Exception exx) {}
			try {
				socket.close();
			} catch (Exception exx) {}
			socket = null;
			dout = null;
			throw ex;
		}
	}

	public synchronized void send(ReflectorPacket packet)
			throws IOException {
		if (logger.isLoggable(Level.FINE))
			logger.fine("SEND=" + packet);
		byte[] buf = packet.getBuffer();
		if (packet.getSocket() != null) {
			// TCP packet
			if (socket != null) {
				try {
					sendTCP(packet);
					return;
				} catch (Exception ex) {}
			}
			DatagramPacket dp = new DatagramPacket(buf, buf.length,
				udpAddress, udpPort);
			dSocket.send(dp);
		} else {
			// UDP packet
			if (udpAddress != null) {
				DatagramPacket dp = new DatagramPacket(buf, buf.length,
					udpAddress, udpPort);
				dSocket.send(dp);
				return;
			}
			sendTCP(packet);
		}
	}

	public synchronized void close() {
		try {
			dout.close();
		} catch (Exception ex) {}
		try {
			socket.close();
		} catch (Exception ex) {}
	}

}
